#version 120

/*
DRAWBUFFERS:01

const float sunPathRotation          =  -41.7; // Sun path rotation on June 20th based in San Francisco.
const float shadowIntervalSize       =    0.5; // Every half a block in movement will update the shadow map.
const float shadowDistanceRenderMul  =    1.0;

const int noiseTextureResolution     =    1024;

const int colortex0Format = RGB10_A2;
const int colortex1Format = RGB16F;
const int colortex2Format = RGB16;
const int colortex3Format = RGBA16;

const bool colortex0Clear = false;
const bool colortex1Clear = false;
const bool colortex2Clear = false;
const bool colortex3Clear = false;

const bool colortex0MipmapEnabled   = true;
const bool colortex1MipmapEnabled   = true;
const bool colortex2MipmapEnabled   = true;

const bool shadowtex0Mipmap         = true;

const bool shadowHardwareFiltering0 = false;
const bool shadowHardwareFiltering1 = true;
*/

uniform vec3 upPosition;
uniform vec3 sunPosition;
uniform vec3 cameraPosition;
uniform vec3 shadowLightPosition;

uniform mat4 shadowModelView;
uniform mat4 gbufferModelView;
uniform mat4 shadowProjection;
uniform mat4 gbufferProjection;
uniform mat4 gbufferModelViewInverse;
uniform mat4 gbufferProjectionInverse;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D depthtex0;
uniform sampler2D depthtex1;
uniform sampler2D shadowtex0;
uniform sampler2D shadowcolor0;

uniform sampler2DShadow shadowtex1;

uniform float sunAngle;
uniform float frameTime;
uniform float viewWidth;
uniform float viewHeight;
uniform float aspectRatio;
uniform float nightVision;
uniform float rainStrength;
uniform float frameTimeCounter;

uniform int isEyeInWater;
uniform int frameCounter;

#define UP_DIRECTION
#define LIGHT_MATRIX
#define SHADOW_MATRIX
#define FOV_PROJECTION

#include "/lib/utilities.glsl"

struct positionStruct {

    mat3x3 viewSpacePosition;
    mat3x3 worldSpacePosition;
    mat2x3 shadowSpacePosition;

} Position;

struct inputStruct {

    vec3 colorTexture1;
    vec3 colorTexture2;

    float depthTexture0;
    float depthTexture1;

} Input;

struct lightingStruct {
    
    vec3 lightShading;
    vec3 indirectLighting;

    mat2x3 lightmapModel;

    float shadingModel;
    float rays;

} Lighting;

struct surfaceStruct {

    float shadedSurface;
    float cloudSurface;

} Surface;

#include "/lib/asset/shadow.glsl"

#include "/lib/utility/misc/ditherIndexMap.glsl"

#include "/lib/utility/sky/computeSky.glsl"
#include "/lib/utility/sky/computeStars.glsl"
#include "/lib/utility/sky/computeClouds2D.glsl"
#include "/lib/utility/sky/computeClouds3D.glsl"

void main() {

    Input.colorTexture1 = textureLoad(colortex1, texCoord).rgb;
    Input.colorTexture2 = textureLoad(colortex2, texCoord).rgb;

    Input.depthTexture0 = textureLoad(depthtex0, texCoord).r;
    Input.depthTexture1 = textureLoad(depthtex1, texCoord).r;

    Position.viewSpacePosition[1] = projectViewSpace(texCoord, Input.depthTexture1, FOV, gbufferProjectionInverse);

    Position.worldSpacePosition[1] = transMAD(gbufferModelViewInverse, Position.viewSpacePosition[1]);

    ivec2 IDSource = decodeIDs(Input.colorTexture2.b);

    int IDs = IDSource.x;
    
    bool hand = bool(IDSource.y);

	vec3 output0 = vec3(0.0);

    Input.colorTexture1 = from16F(Input.colorTexture1);

    output0 = linear(YCoCgToRGB(vec3(Input.colorTexture1.r, decode2x8(Input.colorTexture1.g))));
    
    float curve = 1.0;

    #ifdef ADAPTIVE_SCALING
        curve = clamp01((FPS_TARGET * 0.75) * frameTime + 0.75) / 0.75;
    #endif

    vec2 compositeCoord = (texCoord * COMPOSITE_RES) / curve;

    vec3 tone = computeSky(vec3(0.0));
    
    tone = inversesqrt(vec3(245, 243, 206)) * fInverseLength(tone * 255 + 1.0) + tone;

    Lighting.rays = 0.0;

    Surface.shadedSurface = 0.0;
    Surface.cloudSurface = 0.0;
    
    Lighting.lightmapModel[0] = vec3(0.0);

	#ifdef ADVANCED_FILTERING
        float imbalance = abs(dither - 0.5) * 2.0 + 1.0;

		float variableLOD = (imbalance * 0.5 + 0.5) / COMPOSITE_RES;

        vec2 ditherOffset = imbalance * vec2(1.0, aspectRatio) * phi;
	#else
	 	const float staticLOD = 2.0 / COMPOSITE_RES;
	#endif

    if (!hand) {   
                
        #if CREPUSCULAR_RAYS > 0
            #ifdef ADVANCED_FILTERING
                Lighting.rays = textureLoadLOD(colortex1, vec2( ditherOffset.x,  ditherOffset.y) * screenPixelsSize + compositeCoord, variableLOD).b * 0.25 + Lighting.rays;
                Lighting.rays = textureLoadLOD(colortex1, vec2(-ditherOffset.x,  ditherOffset.y) * screenPixelsSize + compositeCoord, variableLOD).b * 0.25 + Lighting.rays;
                Lighting.rays = textureLoadLOD(colortex1, vec2( ditherOffset.x, -ditherOffset.y) * screenPixelsSize + compositeCoord, variableLOD).b * 0.25 + Lighting.rays;
                Lighting.rays = textureLoadLOD(colortex1, vec2(-ditherOffset.x, -ditherOffset.y) * screenPixelsSize + compositeCoord, variableLOD).b * 0.25 + Lighting.rays;
            #else
                Lighting.rays = textureLoadLOD(colortex1, compositeCoord, staticLOD).b;
            #endif
        #endif

        Lighting.rays = from16F(Lighting.rays);
        
    }
    
    Position.viewSpacePosition[2] = fNormalize(Position.viewSpacePosition[1]);
    Position.worldSpacePosition[2] = fNormalize(Position.worldSpacePosition[1]);

    vec2 lightAura = vec2(0.0);

    lightAura.x = clamp01(dot(Position.viewSpacePosition[2], lightPosition[1]));
    lightAura.y = pow4(lightAura.x);

    if (land) {

        Position.viewSpacePosition[0] = projectViewSpace(texCoord, Input.depthTexture0, FOV, gbufferProjectionInverse);
        
        Position.worldSpacePosition[0] = transMAD(gbufferModelViewInverse, Position.viewSpacePosition[0]);

		Position.shadowSpacePosition[1] = distortShadowSpace(transMAD(shadowProjMatrix, Position.worldSpacePosition[1]), shadowDistanceSize, inverseDistanceSize) * 0.5 + 0.5;

        Lighting.indirectLighting = linear(tone);

        #ifdef ADVANCED_FILTERING
            Lighting.indirectLighting = vec3(0.0);

            Lighting.indirectLighting = textureLoadLOD(colortex0, vec2( ditherOffset.x,  ditherOffset.y) * screenPixelsSize + compositeCoord, variableLOD).rgb * 0.25 + Lighting.indirectLighting;
            Lighting.indirectLighting = textureLoadLOD(colortex0, vec2(-ditherOffset.x,  ditherOffset.y) * screenPixelsSize + compositeCoord, variableLOD).rgb * 0.25 + Lighting.indirectLighting;
            Lighting.indirectLighting = textureLoadLOD(colortex0, vec2( ditherOffset.x, -ditherOffset.y) * screenPixelsSize + compositeCoord, variableLOD).rgb * 0.25 + Lighting.indirectLighting;
            Lighting.indirectLighting = textureLoadLOD(colortex0, vec2(-ditherOffset.x, -ditherOffset.y) * screenPixelsSize + compositeCoord, variableLOD).rgb * 0.25 + Lighting.indirectLighting;
        #else
            Lighting.indirectLighting = textureLoadLOD(colortex0, compositeCoord, staticLOD).rgb;
        #endif

        bool shadowRenderZone = fLength(floor(Position.viewSpacePosition[0] * vec3(16, 16, 256)) / vec3(16, 16, 256)) <= shadowDistance;

		vec3 normal = mat3x3(gbufferModelView) * decodeNormal(Input.colorTexture2.x);

        Lighting.shadingModel = clamp01(dot(lightPosition[1], normal));

        float zCorrect = Lighting.shadingModel > 0.0 ? shadowPixelsSize * (fLength(Position.viewSpacePosition[0]) / shadowDistance) + (shadowPixelsSize * zScaleCoeff / Lighting.shadingModel) : 0.0;

        Position.shadowSpacePosition[1].z -= zCorrect;

        float absorbedLight = shadowRenderZone ? 0.25 : fDistance(Position.shadowSpacePosition[1], vec3(Position.shadowSpacePosition[1].xy, textureLoad(shadowtex0, Position.shadowSpacePosition[1].xy).r));

        if (plant || glass || water)

        Lighting.shadingModel = absorbedLight + 0.5;

        Lighting.lightShading = vec3(0.0);

        float shadowLayers = 0.0;

        if (shadowRenderZone) {
                
            shadowLayers = textureLoad(shadowcolor0, Position.shadowSpacePosition[1].xy).a;

            Surface.shadedSurface = shadowLoad(shadowtex1, Position.shadowSpacePosition[1]).r;

                #if SHADOW_FILTER >= 1
                    vec4 blurSize = vec4(shadowPixelsSize, shadowPixelsSize, 0.0, 0.0) * (plant ? absorbedLight + 2.0 * 2.0 : 0.5);

                    const float sampleRCP = RCP(SHADOW_FILTER_SAMPLES);

                    float ditherScale = 0.0;

                    #if SHADOW_FILTER == 2
                        const float softnessRCP = 0.25;

                        const float heightSampleScale = 256.0 * softnessRCP;

                        ditherScale = dither * 4.0 * softnessRCP;

                        float distanceLOD = 4 * fInverseLength(Position.viewSpacePosition[2] + 1.0);

                        for (int i = 0; i < 4; ++i)
                        blurSize.w = (Position.shadowSpacePosition[1].z - textureLoadLOD(shadowtex0, hammersleyPointMap(i, softnessRCP * i + 1.0, ditherScale) * blurSize.xy + Position.shadowSpacePosition[1].xy, distanceLOD).r) * heightSampleScale + blurSize.w;
                        
                        blurSize.w = abs(blurSize.w) / sunAngularDiameter;

                        blurSize.xyz = blurSize.w * blurSize.xyz + blurSize.xyz;
                    #endif

                    ditherScale = dither * 16 * sampleRCP;

                    for (int i = 0; i < SHADOW_FILTER_SAMPLES; ++i)
                    Lighting.lightShading = shadowLoad(shadowtex1, vec3(hammersleyPointMap(i, sampleRCP * i + 1.0, ditherScale), 0.0) * blurSize.xyz + Position.shadowSpacePosition[1]).r * sampleRCP + Lighting.lightShading;
                #else 
                    Lighting.lightShading += shadowLoad(shadowtex1, Position.shadowSpacePosition[1]).r;
                #endif

                #ifdef SHADOW_COLORED_SHADOWS  
                    vec3 coloredShading = vec3(0.0);

                    #if SHADOW_FILTER >= 1
                        for (int i = 0; i < SHADOW_FILTER_SAMPLES; ++i) {
                            
                            vec3 currentPos = vec3(hammersleyPointMap(i, sampleRCP * i + 1.0, ditherScale), 0.0) * shadowPixelsSize + Position.shadowSpacePosition[1];

                            if (abs(textureLoad(shadowcolor0, currentPos.xy).a - 0.5) < 0.25) 

                            coloredShading = textureLoad(shadowcolor0, currentPos.xy).rgb * sampleRCP + coloredShading;

                        }
                    #else
                        if (abs(textureLoad(shadowcolor0, Position.shadowSpacePosition[1].xy).a - 0.5) < 0.25) 
                        
                        coloredShading = textureLoad(shadowcolor0, Position.shadowSpacePosition[1].xy).rgb;
                    #endif

                    Lighting.lightShading = mix(Lighting.lightShading, coloredShading * Lighting.lightShading * tone, clamp01(fLength(coloredShading)));
                #endif


        } else 

        Lighting.lightShading = vec3(0.0);

        Lighting.lightmapModel[0].xy = decode2x8(Input.colorTexture2.y);

        Lighting.lightmapModel[1] = sqrt(pow8(Lighting.lightmapModel[0].x) * sqrt(emissiveColor * fExp(-lumaCoeff)));

        Lighting.indirectLighting = max(sqrt(Lighting.indirectLighting), rcp8bit) + Lighting.lightmapModel[1];

        float luminance = luma(output0);

        float saturationLevel = fLength(Lighting.indirectLighting * fInverseLength(Lighting.indirectLighting + RCP(sqrt(sunLuminance)))) / (luminance * exp2(saturationCoeffLn * desaturationCoeff) + saturationCoeff);

        output0 = ((output0 - luminance) * saturationLevel + luminance) * Lighting.lightmapModel[1];

        vec3 fogLight = fExp(-waterColor.rgb) * waterColor.a * tone * Lighting.rays * Lighting.lightmapModel[1];

        if (water || underWater) {
            
            Position.shadowSpacePosition[0] = distortShadowSpace(transMAD(shadowProjMatrix, Position.worldSpacePosition[0]), shadowDistanceSize, inverseDistanceSize) * 0.5 + 0.5;

            Position.shadowSpacePosition[0].z -= zCorrect;

            Surface.shadedSurface = shadowLoad(shadowtex1, Position.shadowSpacePosition[0]).r;

            if (!underWater) {

                #ifdef WATER_FOG
                    vec4 accumulateFog = vec4(0.0);

                    accumulateFog.xyz = (Position.viewSpacePosition[1] - Position.viewSpacePosition[0]);
                    
                    accumulateFog.w = (fLength(accumulateFog.xyz) * abs(dot(normal, fNormalize(accumulateFog.xyz))));
                    
                    accumulateFog.w = clamp01(sqrt(abs(fExp(-accumulateFog.w / waterColor.a))));
          
                    output0 *= exp2(-waterColor.rgb * (accumulateFog.w * waterColor.a) - waterColor.rgb) * fogLight + 0.5;

                    output0 = mix(accumulateFog.w * fogLight, output0, accumulateFog.w);
                #endif
                
            }
            
        }

        #ifdef WATER_FOG
            if (underWater) output0 = mix(fogLight, output0 * exp2(-waterColor.rgb), fExp(-fLength(Position.viewSpacePosition[0])));
        #endif

    } else {

        output0 = computeSky(Position.viewSpacePosition[2]) * output0;

        vec2 horizon = vec2(0.0);

        horizon.x = pow8(lightAura.y) * 0.25 + clamp01(dot(Position.viewSpacePosition[2], upDirection));

        if (horizon.x > 0.1) {

            horizon.y = min(pow2(horizon.x), 0.25) * horizon.x * inverseCoverage * mix(lightAura.y * 0.75 + 0.25, 1.0, fLength(tone));

            float cloudMovement = time * 0.02;

            float yRCP = RCP(Position.worldSpacePosition[2].y);

            #if CLOUDS == 3 || CLOUDS == 4
                computeThreeDimensionalClouds(output0, Position.worldSpacePosition[2], yRCP, tone, horizon, lightAura, cloudMovement, Surface.cloudSurface);
            #endif

            #if CLOUDS == 2 || CLOUDS == 4
                computeTwoDimensionalClouds(output0, Position.worldSpacePosition[2], tone, horizon, lightAura, cloudMovement, Surface.cloudSurface);
            #endif

            #ifdef STARS
                computeStars(output0, Position.worldSpacePosition[2], horizon, Surface.cloudSurface);
            #endif

        }
    }
    
    output0 = Lighting.rays * (tone * (lightAura.y * rcpln2 + ln2)) + sqrt(abs(output0));

    output0 = abs((dither - 0.5) * rcp16bit + output0);
    
    gl_FragData[0] = vec4(Surface.shadedSurface, Surface.cloudSurface, Lighting.lightmapModel[0].y > 0.75, 1.0);
    gl_FragData[1] = vec4(to16F(output0), 1.0);
}