#version 120

/*
DRAWBUFFERS:12

const bool colortex0MipmapEnabled = true;
const bool colortex1MipmapEnabled = true;
*/

uniform vec3 upPosition;
uniform vec3 sunPosition;
uniform vec3 cameraPosition;
uniform vec3 shadowLightPosition;

uniform sampler2D noisetex;
uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D depthtex0;
uniform sampler2D depthtex1;

uniform mat4 gbufferModelView;
uniform mat4 gbufferProjection;
uniform mat4 gbufferModelViewInverse;
uniform mat4 gbufferProjectionInverse;

uniform float sunAngle;
uniform float frameTime;
uniform float viewWidth;
uniform float viewHeight;
uniform float rainStrength;
uniform float frameTimeCounter;

uniform int isEyeInWater;
uniform int frameCounter;

#define UP_DIRECTION
#define LIGHT_MATRIX
#define FOV_PROJECTION

#include "/lib/utilities.glsl"

mat2x3 reflectionMatrix = mat2x3(0.0);

struct inputStruct {

    vec3 colorTexture0;
    vec3 colorTexture1;
    vec3 colorTexture2;

    float depthTexture0;
    float depthTexture1;

} Input;

struct positionStruct {

	mat3x3 viewSpacePosition;

	vec3 worldSpacePosition;

	mat2x3 reflectedRay;

} Position;

#include "/lib/utility/misc/ditherIndexMap.glsl"

#include "/lib/utility/sky/computeSky.glsl"
#include "/lib/utility/sky/computeStars.glsl"
#include "/lib/utility/sky/computeClouds2D.glsl"
#include "/lib/utility/sky/computeClouds3D.glsl"

#include "/lib/utility/water/waterWaves.glsl"
#include "/lib/utility/water/waterParallax.glsl"
#include "/lib/utility/water/computeRaytracedCoordinate.glsl"

#include "/lib/utility/refract/computeRefractionComposite2.glsl"

void main() {

    Input.colorTexture1 = textureLoad(colortex1, texCoord).rgb;
    Input.colorTexture2 = textureLoad(colortex2, texCoord).rgb;

    Input.depthTexture0 = textureLoad(depthtex0, texCoord).r;

    Position.viewSpacePosition[0] = projectViewSpace(texCoord, Input.depthTexture0, FOV, gbufferProjectionInverse);

    Position.viewSpacePosition[2] = fNormalize(Position.viewSpacePosition[0]);

    Position.reflectedRay = mat2x3(0.0);

    ivec2 IDSource = decodeIDs(Input.colorTexture2.b);

    int IDs = IDSource.x;

    bool hand = bool(IDSource.y);

    vec3 output0 = linear(from16F(Input.colorTexture1));
    
    float sky = 0.0;
    float reflectedIDs = 0.0;

    vec3 tone = computeSky(vec3(0.0));

    tone = inversesqrt(vec3(245, 243, 206)) * fInverseLength(tone * 255 + 1.0) + tone;

	if (underWater) {

        #ifdef REFRACTION
            #ifdef REFRACTION_UNDERWATER_DISTORTION
                computeUnderwaterRefraction(output0, Input.colorTexture2);
            #endif
        #endif

    }

    if (!land) {
        
        sky = step(sunAngularDiameterCos, dot(Position.viewSpacePosition[2], lightPosition[0])) * sunLuminance + sky;
        sky = step(moonAngularDiameterCos, dot(Position.viewSpacePosition[2], -lightPosition[0])) * moonLuminance + sky;
             
        sky = sky * (1.0 - textureLoadLOD(colortex0, texCoord, 0.125 + dither).g) * clamp01(dot(Position.viewSpacePosition[0], upDirection));

        output0 = linear(sqrt(sky) * tone + from16F(textureLoadLOD(colortex1, texCoord, 0.125 + dither).rgb));

    }

    if (water) {

        Input.depthTexture1 = textureLoad(depthtex1, texCoord).r;

        Position.viewSpacePosition[1] = projectViewSpace(texCoord, Input.depthTexture1, 1.0, gbufferProjectionInverse);

        Position.worldSpacePosition = transMAD(gbufferModelViewInverse, Position.viewSpacePosition[0]);

        vec3 normal = mat3(gbufferModelView) * decodeNormal(Input.colorTexture2.r);

        mat2x3 waves = mat2x3(normal, normal);

        #if defined WATER_NORMALS || defined REFRACTION_WATER_WAVES
            vec3 tangent = cross(normal, gbufferModelView[2].xyz);
            
            mat3 tbnMatrix = transpose(mat3(tangent, cross(tangent, normal), normal));

            mat2x3 position = mat2x3(0.0);

            position[0] = Position.worldSpacePosition + cameraPosition;
            position[1] = tbnMatrix * Position.viewSpacePosition[2];

            #ifdef WATER_PARALLAX_WAVES
                getParallaxWaves(position);
            #endif

            waves[0].y = getWaves(position[0]);
            waves[0].x = getWaves(position[0] + vec3(wavesDifferentialCoeff, 0.0, 0.0));
            waves[0].z = getWaves(position[0] + vec3(0.0, 0.0, wavesDifferentialCoeff));

            waves[1].xy = (waves[0].y - waves[0].xz) / wavesDifferentialCoeff;

            waves[1].z = sqrt(abs(1.0 - dotX(waves[1].xy)));

            waves[0] = waves[1];
            waves[1] = waves[1] * tbnMatrix;
        #endif

        #ifdef REFRACTION 
            #ifdef REFRACTION_WATER_WAVES
                computeWaterRefraction(output0, waves, Position);
            #endif
        #endif
        
        #ifdef WATER_REFLECTIONS
            #define rayCoord reflectionMatrix[0]
            #define reflection reflectionMatrix[1]

            Position.viewSpacePosition[1] = Position.viewSpacePosition[0];

            Position.reflectedRay[0] = reflect(Position.viewSpacePosition[2], waves[1]);
            Position.reflectedRay[1] = mat3(gbufferModelViewInverse) * Position.reflectedRay[0];

            float fresnel = pow5(dot(waves[1], Position.viewSpacePosition[2]) * 0.25 + 0.75);
 
            float cloudSurface = 0.0;

            int refinements = 0;

            bool hit = false;
            
            const float sampleRCP = RCP(WATER_ANISOTROPIC_SAMPLES);

            vec2 stretchSize = sampleRCP * fresnel * screenPixelsSize * ivec2(4, 32) * (dither * 8 * sampleRCP);
            
            for (int i = 0; i < WATER_REFLECTION_LAYERS; ++i) {

                refinements = 0;

                hit = false;

                computeRaytracedCoordinates(rayCoord, hit, refinements, Position);

                if (hit) {

                    #if WATER_REFLECTION_TYPE == 0
                        reflection += linear(from16F(textureLoad(colortex1, rayCoord.xy).rgb * rcp16bit));
                    #else
                        for (int i = 0; i < WATER_ANISOTROPIC_SAMPLES; ++i)
                        reflection = linear(from16F(textureLoad(colortex1, rotatePoint(i) * stretchSize + rayCoord.xy).rgb)) * sampleRCP + reflection;
                    #endif

                    Input.colorTexture2 = textureLoad(colortex2, rayCoord.xy).rgb;

                    reflectedIDs = decodeIDs(Input.colorTexture2.b).x;
                    
                }

                if (!waterCheck(reflectedIDs)) break;
                if (i == WATER_REFLECTION_LAYERS) break;

                Input.depthTexture0 = textureLoad(depthtex0, rayCoord.xy).r;
                Input.depthTexture1 = textureLoad(depthtex1, rayCoord.xy).r;

                Position.viewSpacePosition[0] = projectViewSpace(rayCoord.xy, Input.depthTexture0, FOV, gbufferProjectionInverse);
                Position.viewSpacePosition[1] = projectViewSpace(rayCoord.xy, Input.depthTexture1, 1.0, gbufferProjectionInverse);

                Position.viewSpacePosition[2] = fNormalize(Position.viewSpacePosition[0]);

                Position.worldSpacePosition = transMAD(gbufferModelViewInverse, Position.viewSpacePosition[0]);

                normal = mat3(gbufferModelView) * decodeNormal(Input.colorTexture2.r);

                waves = mat2x3(normal, normal);
                
                #if defined WATER_NORMALS
                    vec3 tangent = cross(normal, gbufferModelView[2].xyz);
                    
                    mat3 tbnMatrix = transpose(mat3(tangent, cross(tangent, normal), normal));

                    mat2x3 position = mat2x3(0.0);

                    position[0] = Position.worldSpacePosition + cameraPosition;
                    position[1] = tbnMatrix * Position.viewSpacePosition[2];

                    #ifdef WATER_PARALLAX_WAVES
                        getParallaxWaves(position);
                    #endif

                    waves[0].y = getWaves(position[0]);
                    waves[0].x = getWaves(position[0] + vec3(wavesDifferentialCoeff, 0.0, 0.0));
                    waves[0].z = getWaves(position[0] + vec3(0.0, 0.0, wavesDifferentialCoeff));

                    waves[1].xy = (waves[0].y - waves[0].xz) / wavesDifferentialCoeff;

                    waves[1].z = sqrt(abs(1.0 - dotX(waves[1].xy)));

                    waves[1] = waves[1] * tbnMatrix;
                #endif
                
                Position.viewSpacePosition[1] = Position.viewSpacePosition[0];

                Position.reflectedRay[0] = reflect(Position.viewSpacePosition[2], waves[1]);
                Position.reflectedRay[1] = mat3(gbufferModelViewInverse) * Position.reflectedRay[0];

            }

            if (textureLoad(colortex0, texCoord).b > 0.0 && reflectedIDs == 0) {

                reflection = computeSky(Position.reflectedRay[0]);

                vec2 lightAura = vec2(0.0);
            
                lightAura.x = clamp01(dot(Position.reflectedRay[0], lightPosition[1]));
                lightAura.y = pow4(lightAura.x);
                
                vec2 horizon = vec2(0.0);
            
                horizon.x = pow8(lightAura.y) * 0.25 + clamp01(dot(Position.reflectedRay[0], upDirection)) * mix(lightAura.y * 0.75 + 0.25, 1.0, fLength(tone));

                if (horizon.x > 0.1) {

                    horizon.y = min(pow2(horizon.x), 0.25) * horizon.x * inverseCoverage;

                    float cloudMovement = time * 0.02;

                    float yRCP = RCP(Position.reflectedRay[1].y);

                    #if CLOUDS == 3 || CLOUDS == 4
                        computeThreeDimensionalClouds(reflection, Position.reflectedRay[1], yRCP, tone, horizon, lightAura, cloudMovement, cloudSurface);
                    #endif

                    #if CLOUDS == 2 || CLOUDS == 4
                        computeTwoDimensionalClouds(reflection, Position.reflectedRay[1], tone, horizon, lightAura, cloudMovement, cloudSurface);
                    #endif

                    #ifdef STARS
                        computeStars(reflection, Position.reflectedRay[1], horizon, cloudSurface);
                    #endif

                }

                #if CREPUSCULAR_RAYS > 0
                    reflection = linear(0.5 * tone * (lightAura.y * rcpln2 + ln2)) + reflection;
                #endif
            }

            if (textureLoad(colortex0, texCoord).r > 0.0 && reflectedIDs == 0) {

                sky = step(sunAngularDiameterCos, dot(Position.reflectedRay[0], lightPosition[0])) * sunLuminance + sky;
                sky = step(moonAngularDiameterCos, dot(Position.reflectedRay[0], -lightPosition[0])) * moonLuminance + sky;

                sky = sky * (1.0 - cloudSurface);

            }

            output0 = mix(sky * linear(tone) + output0, reflection, fresnel);
        #endif

    }

    output0 = sqrt(abs(output0));
    
    if (emissive) output0 = output0 / dotX(lumaCoeff * 0.25 + 0.25) + output0;

    gl_FragData[0] = vec4(output0, 1.0);
    gl_FragData[1] = vec4(warm, hand, textureLoad(noisetex, texCoord).r, 1.0);
}